/*
 * Esta classe contem os metodos utilizados para realizar a classificacao
 * de obras entre compositores.
 */

package classificador;

import aritmetico.ArithCodeOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;
import ppm.PPMModel;

/**
 *
 * @author dj
 */
public class Classificador {

    private LinkedList<Compositor> listaCompositores = new LinkedList<Compositor>();
    static int tempFileNumber = 0;
    private LinkedList<Resultado> listaResultadosAV1M = new LinkedList<Resultado>(); //Avaliacao V1M
    private LinkedList<Resultado> listaResultadosAV1R = new LinkedList<Resultado>(); //Avaliacao V1R
    private LinkedList<Resultado> listaResultadosAV2 = new LinkedList<Resultado>(); //Avaliacao V2
    private LinkedList<Resultado> listaResultadosAV3 = new LinkedList<Resultado>(); //Avaliacao V3
    private LinkedList<Resultado> listaResultadosAV4 = new LinkedList<Resultado>(); //Avaliacao V4

    /**
     * Diretorio padrao para os arquivos temporarios gerados
     */
    private File diretorioTemp = new File("diretoriox");

    public Classificador(){
        boolean mkdir = diretorioTemp.mkdir();
    }

    /**
     * Adiciona um novo PPMModel ao classificador
     */
    public void adicionarCompositor(){
        adicionarCompositor("Compositor "+(1+listaCompositores.size()));
    }

    /**
     * Adiciona um novo Compositor ao classificador utilizando um nome especifico.
     * @param nomeCompositor
     */
    public void adicionarCompositor(String nomeCompositor){
        listaCompositores.add(new Compositor(nomeCompositor));
    }

    /**
     * Prepara os arquivos dos Compositores da listaCompositores para iniciar 
     * uma classificacao
     */
    public void prepararArquivosDosCompositores(){
        for(Compositor compositor: listaCompositores){
            for(Obra arquivo: compositor.getListaDeArquivos()){
                arquivo.preencherListaDeNotas();
                arquivo.gerarListaDeArticulacoes(0);
                arquivo.gerarListaDeArticulacoes(1);
                arquivo.gerarListaDeArticulacoes(2);
            }
        }
    }

    /**
     * Classifica uma obra entre todos os Compositores ativos comparando as taxas
     * de compressao a fim de identificar o autor.
     * @param obra Obra a ser classificada entre os Compositores
     * @return Retorna o ID do compositor que melhor comprimiu a obra
     * @throws IOException
     */
    public int classificarArquivo(File obra, int tipoDeModelo) throws IOException{
        Obra arquivoAClassificar = new Obra(obra);
        int compositorMaisAdequado = Integer.MAX_VALUE;
        double razaoCompressaoMin = Double.MAX_VALUE;
        double razaoCompressaoTemp = Double.MAX_VALUE;

        if(getListaCompositores().size() > 0){
            arquivoAClassificar.preencherListaDeNotas();
            System.out.println(" ");
            prepararArquivosDosCompositores();
            for(int i = 0; i < getListaCompositores().size(); i++){
                if(getListaCompositores().get(i).isAtivo() && getListaCompositores().get(i).isComArquivos()){
                    getListaCompositores().get(i).gerarModeloPPM(tipoDeModelo);
                    razaoCompressaoTemp = comprimeObraComModeloDoCompositor(arquivoAClassificar,
                                                             getListaCompositores().get(i), tipoDeModelo);
                    if(razaoCompressaoTemp < razaoCompressaoMin){
                        razaoCompressaoMin = razaoCompressaoTemp;
                        compositorMaisAdequado = i;
                    }
                }
            }
            //se nenhum compositor estiver ativo ou populado, retornar -1
            if(compositorMaisAdequado == Integer.MAX_VALUE) compositorMaisAdequado = -1;
        }
        else{//se nao existir compositor, retornar -1
            compositorMaisAdequado = -1;
        }
        return compositorMaisAdequado;
    }

    /**
     * Metodo para realizar a compressao de uma obra baseando-se em um
     * compositor e utilizando o PPM-C. O retorno sera a taxa de
     * compressao em porcentagem.
     * @param obra Obra a ser comprimida
     * @param compositor PPMModel a ser utilizado para a compressao
     * @param tipoDeModelo Tipo de modelo a ser utilizado: 0 = M+R, 1 = M, 2 = R
     * @return Retorna a porcentagem de compressao
     * @throws FileNotFoundException
     * @throws IOException
     */
    public double comprimeObraComModeloDoCompositor(Obra obra, Compositor compositor, int tipoDeModelo) throws FileNotFoundException, IOException{

        double razaoDeCompressao = 0;
        tempFileNumber++;
        File arquivoTemp = File.createTempFile("temp".concat(Integer.toString(tempFileNumber)), ".ppmmdd", diretorioTemp);
 
        PPMModel modelo = compositor.getModelo(tipoDeModelo);
                
        ArithCodeOutputStream codigoSaida = new ArithCodeOutputStream(new FileOutputStream(arquivoTemp), modelo, false);

        if(!obra.isComNotas()){
            obra.preencherListaDeNotas();
        }
        
        LinkedList<Integer> articulacao;
        obra.gerarListaDeArticulacoes(tipoDeModelo) ;
        articulacao = obra.getListaDeArticulacoes(tipoDeModelo);

        for(Integer intervalo : articulacao){
            codigoSaida.encode(intervalo, false);
        }
        
        codigoSaida.close();

        razaoDeCompressao =  (double) ((double) arquivoTemp.length())/((double)articulacao.size());

        System.out.println("  -Obra: "+obra.getArquivoMidi().getName());
        System.out.println("  -Modelo do Compositor: "+compositor.getNome());
        System.out.println("  -Razao de compressao: "+arquivoTemp.length() +" / "+articulacao.size()+" = "+razaoDeCompressao+"%\n");

        arquivoTemp.deleteOnExit();
        
        return razaoDeCompressao;
    }

    /**
     * Metodo para calcular o limite minimo de intervalos existente nas melodias
     * de todos os listaCompositores
     * @return limite
     */
    public int calcularLimite(){
        int limite = Integer.MAX_VALUE;
        int tempLimite = 0;

        for (Compositor compositor : listaCompositores) {
            tempLimite = compositor.checarMenorQuantidadeDeIntervalos();
            if(limite > tempLimite){
                limite = tempLimite;
            }    
        }
        
        return limite;
    }

    /**
     * AV1 - classifica arquivos baseando-se em melodia ou ritmo, classificando
     * a partir da Razao de Compressao em cada modelo.
     * 
     * Realiza a avaliacao do arquivo utilizando o PPM e considerando cada arquivo
     * da base de dados como um modelo a fim de encontrar os mais semelhantes.
     * O contextoInicial e contextoFinal serve para programar as avaliacoes 
     * que serao realizadas. Para que seja realizada apenas uma avaliacao em um 
     * contexto especifico, o contexto inicial deve ser igual ao final. 
     * 
     * Tipo de Modelo:
     *  rit+mel = 0 (ainda nao implementado)
     *  mel = 1
     *  rit = 2
     *
     * @param contextoInicial
     * @param contextoFinal
     * @param tipoDeModelo
     * @return
     * @throws IOException
     */
    public boolean avaliacaoV1(int contextoInicial, int contextoFinal, int tipoDeModelo) throws IOException{
        if(listaCompositores.size() == 0){
            System.out.println("Nao existem Compositores para realizar a classificacao");
            return false;
        }
        
        boolean compositoresComModelosEqualizados = true;
        int quantidadeDeArquivos = listaCompositores.getFirst().getListaDeArquivos().size();
        int compositoresAtivos = 0;
        Obra[] obrasAClassificar;
        String[] nomeCompositores;
        int tipoAtual = tipoDeModelo;
        
        for(int i = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                compositoresAtivos++;
                if(listaCompositores.get(i).getListaDeArquivos().size() != quantidadeDeArquivos){
                    compositoresComModelosEqualizados = false;
                    System.out.println("Compositor "+listaCompositores.get(i).getNome()+" apresenta-se com quantidade diferente de arquivos!");
                }//fim if listaCompositores completos
            }//fim if listaCompositores ativos
        }//fim for lista de arquivos

        if(compositoresAtivos == 0 || compositoresComModelosEqualizados == false){
            System.out.println("Nao foi possivel classificar: Nenhum compositor ativo ou algum modelo esta incompativel!");
            return false;
        }
        
        obrasAClassificar = new Obra[compositoresAtivos];
        nomeCompositores = new String[compositoresAtivos];

        for(int i = 0, j = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                nomeCompositores[j] = Integer.toString(j).concat(":").concat(listaCompositores.get(i).getNome());
                j++;
            }
        }
        
        if(contextoInicial < 0){
            contextoInicial = 0;
        }
        if(contextoFinal < contextoInicial){
            contextoFinal = contextoInicial;
        }

         // TODO: Falta implementar o modo Melodia+Ritmo
        if(tipoDeModelo == 1){
            listaResultadosAV1M = new LinkedList<Resultado>();
        }else if (tipoDeModelo == 2){
            listaResultadosAV1R = new LinkedList<Resultado>();
        }

        for(int c = contextoInicial; c <= contextoFinal; c++){ //realiza validacao cruzada com contextos determinados
            Compositor.setContexto(c);
            // TODO: Falta implementar o modo Melodia+Ritmo
            Resultado resultadoFinal = new Resultado("Avaliacao_V1",
                                             Compositor.getContexto(),
                                             (tipoAtual == 1), 
                                             (tipoAtual == 2), 
                                             compositoresAtivos, 
                                             quantidadeDeArquivos);
            resultadoFinal.setNomeCompositores(nomeCompositores);

            //inicio do classificador estatistico
            System.out.println("\nClassificacao tipo "+tipoAtual+" contexto "+c);
            for(int j = 0; j < quantidadeDeArquivos; j++){

                //preenche a lista com os arquivos que serao classificados
                for(int i = 0, k = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos()){
                        obrasAClassificar[k] = listaCompositores.get(i).getListaDeArquivos().removeFirst();
                        k++;
                    }
                }

                //classifica cada arquivo em cada compositor
                // a = index arquivo atual
                for(int a = 0; a < obrasAClassificar.length; a++){
                    System.out.println("\nClassificando: "+obrasAClassificar[a].getArquivoMidi().getName());

                    LinkedList<ItemResultado> itensResultadoObraAtual = new LinkedList<ItemResultado>();

                    itensResultadoObraAtual.add(new ItemResultado(a));
                    itensResultadoObraAtual.getLast().setNomeDoArquivo(obrasAClassificar[a].getArquivoMidi().getName());

                    // m = index do compositor atual
                    for(int m = 0; m < listaCompositores.size(); m++){
                        if(listaCompositores.get(m).isAtivo() && listaCompositores.get(m).isComArquivos()){
                            listaCompositores.get(m).gerarModeloPPM(tipoAtual);
                            double taxa = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), tipoAtual);
                            itensResultadoObraAtual.addLast(new ItemResultado(m));
                            itensResultadoObraAtual.getLast().setValor(taxa);
                        }
                    }
                    resultadoFinal.inserirResultado(itensResultadoObraAtual);                        
                }
                for(int i = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos())
                        listaCompositores.get(i).getListaDeArquivos().addLast(obrasAClassificar[i]);
                }
            }
            resultadoFinal.ordenarResultados();
            resultadoFinal.calcularAcertos();
            
            if(tipoDeModelo == 1){
                listaResultadosAV1M.add(resultadoFinal);
            }else if (tipoDeModelo == 2){
                listaResultadosAV1R.add(resultadoFinal);
            }
        }
        return true;
    }

    /**
     * AV2 - classifica arquivos baseando-se em melodia e ritmo, avaliando o melhor
     * resultado a partir da media dos resultados obtidos na avaliacao da melodia 
     * e ritmo de um mesmo modelo em um mesmo contexto.
     * 
     * Realiza a avaliacao do arquivo utilizando o PPM e considerando cada arquivo
     * da base de dados como um modelo a fim de encontrar os mais semelhantes.
     * O contextoInicial e contextoFinal serve para programar as avaliacoes 
     * que serao realizadas. Para que seja realizada apenas uma avaliacao em um 
     * contexto especifico, o contexto inicial deve ser igual ao final. 
     *
     * @param contextoInicial
     * @param contextoFinal
     * @return
     * @throws IOException
     */
    public boolean avaliacaoV2(int contextoInicial, int contextoFinal) throws IOException{
        if(listaCompositores.size() == 0){
            System.out.println("Nao existem Compositores para realizar a classificacao");
            return false;
        }
        
        boolean compositoresComModelosEqualizados = true;
        int quantidadeDeArquivos = listaCompositores.getFirst().getListaDeArquivos().size();
        int compositoresAtivos = 0;
        Obra[] obrasAClassificar;
        String[] nomeCompositores;

        for(int i = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                compositoresAtivos++;
                if(listaCompositores.get(i).getListaDeArquivos().size() != quantidadeDeArquivos){
                    compositoresComModelosEqualizados = false;
                    System.out.println("Compositor "+listaCompositores.get(i).getNome()+" apresenta-se com quantidade diferente de arquivos!");
                }//fim if listaCompositores completos
            }//fim if listaCompositores ativos
        }//fim for lista de arquivos

        if(compositoresAtivos == 0 || compositoresComModelosEqualizados == false){
            System.out.println("Nao foi possivel classificar: Nenhum compositor ativo ou algum modelo incompativel!");
            return false;
        }
        
        obrasAClassificar = new Obra[compositoresAtivos];
        nomeCompositores = new String[compositoresAtivos];
        
        for(int i = 0, j = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                nomeCompositores[j] = Integer.toString(j).concat(":").concat(listaCompositores.get(i).getNome());
                j++;
            }
        }
        
        if(contextoInicial < 0){
            contextoInicial = 0;
        }
        if(contextoFinal < contextoInicial){
            contextoFinal = contextoInicial;
        }

        listaResultadosAV2 = new LinkedList<Resultado>();

        for(int c = contextoInicial; c <= contextoFinal; c++){ //realiza validacao cruzada com contextos determinados
            Compositor.setContexto(c);

            Resultado resultadoFinal = new Resultado("Avaliacao_V2",
                                             Compositor.getContexto(),
                                             true, 
                                             true, 
                                             compositoresAtivos, 
                                             quantidadeDeArquivos);
            resultadoFinal.setNomeCompositores(nomeCompositores);

            System.out.println("\nClassificacao com contexto:"+c+"\n");
            //inicio do classificador
            for(int j = 0; j < quantidadeDeArquivos; j++){
                //preenche a lista com os arquivos que serao classificados
                for(int i = 0, k = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos()){
                        obrasAClassificar[k] = listaCompositores.get(i).getListaDeArquivos().removeFirst();
                        k++;
                    }
                }
                //classifica cada arquivo em cada compositor
                // a = index arquivo atual
                for(int a = 0; a < obrasAClassificar.length; a++){
                    System.out.println("\nClassificando: "+obrasAClassificar[a].getArquivoMidi().getName());
                    LinkedList<ItemResultado> resultadoObraAtual =  new LinkedList<ItemResultado>();
                    resultadoObraAtual.add(new ItemResultado(a));
                    resultadoObraAtual.getLast().setNomeDoArquivo(obrasAClassificar[a].getArquivoMidi().getName());

                    // m = index do compositor atual
                    for(int m = 0; m < listaCompositores.size(); m++){
                        if(listaCompositores.get(m).isAtivo() && listaCompositores.get(m).isComArquivos()){
                            listaCompositores.get(m).gerarModeloPPM(1);
                            double taxaMelodia = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), 1);
                            listaCompositores.get(m).gerarModeloPPM(2);
                            double taxaRitmo = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), 2);

                            resultadoObraAtual.addLast(new ItemResultado(m));
                            resultadoObraAtual.getLast().setValor((taxaMelodia+taxaRitmo)/2.0) ;
                        }
                    }
                    resultadoFinal.inserirResultado(resultadoObraAtual);                        
                }
                for(int i = 0, k = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos())
                        listaCompositores.get(i).getListaDeArquivos().addLast(obrasAClassificar[k]);
                        k++;
                }
            }
            resultadoFinal.ordenarResultados();
            resultadoFinal.calcularAcertos();
            listaResultadosAV2.add(resultadoFinal);
            
        }//fim do for para gerar resultados
        
        
        return true;
    }
    
    /**
     * AV3 - classifica arquivos baseando-se em melodia e ritmo, avaliando o melhor
     * resultado a partir da media das posicoes no ranking obtidas na avaliacao da melodia 
     * e ritmo de um mesmo modelo em um mesmo contexto. Arquivos com a mesma taxa
     * serao considerados na mesma posicao do ranking.
     * 
     * Realiza a avaliacao do arquivo utilizando o PPM e considerando cada arquivo
     * da base de dados como um modelo a fim de encontrar os mais semelhantes.
     * O contextoInicial e contextoFinal serve para programar as avaliacoes 
     * que serao realizadas. Para que seja realizada apenas uma avaliacao em um 
     * contexto especifico, o contexto inicial deve ser igual ao final. 
     *
     * @param contextoInicial
     * @param contextoFinal
     * @return
     * @throws IOException
     */
    public boolean avaliacaoV3(int contextoInicial, int contextoFinal) throws IOException{
        if(listaCompositores.size() == 0){
            System.out.println("Nao existem Compositores para realizar a classificacao");
            return false;
        }
        
        boolean compositoresComModelosEqualizados = true;
        int quantidadeDeArquivosPorCompositor = listaCompositores.getFirst().getListaDeArquivos().size();
        int compositoresAtivos = 0;
        Obra[] obrasAClassificar;
        String[] nomeCompositores;
        
        // verificacao especial para a classificacao v3
        if(quantidadeDeArquivosPorCompositor < 2){
            System.out.println("Nao existem arquivos suficientes para realizar a classificacao");
            return false;
                 
        }

        for(int i = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                compositoresAtivos++;
                if(listaCompositores.get(i).getListaDeArquivos().size() != quantidadeDeArquivosPorCompositor){
                    compositoresComModelosEqualizados = false;
                    System.out.println("Compositor "+listaCompositores.get(i).getNome()+" apresenta-se com quantidade diferente de arquivos!");
                }//fim if listaCompositores completos
            }//fim if listaCompositores ativos
        }//fim for lista de arquivos

        if(compositoresAtivos == 0 || compositoresComModelosEqualizados == false){
            System.out.println("Nao foi possivel classificar: Nenhum compositor ativo ou algum modelo incompativel!");
            return false;
        }
        
        obrasAClassificar = new Obra[compositoresAtivos];
        nomeCompositores = new String[compositoresAtivos];
        
        for(int i = 0, j = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                nomeCompositores[j] = Integer.toString(j).concat(":").concat(listaCompositores.get(i).getNome());
                j++;
            }
        }
        
        if(contextoInicial < 0){
            contextoInicial = 0;
        }
        if(contextoFinal < contextoInicial){
            contextoFinal = contextoInicial;
        }

        listaResultadosAV4 = new LinkedList<Resultado>();
        
        for(int c = contextoInicial; c <= contextoFinal; c++){ //realiza validacao cruzada com contextos determinados
            Compositor.setContexto(c);
            
            Resultado resultadoMelodia = new Resultado("Avaliacao_V3-Melodia",
                                             Compositor.getContexto(),
                                             true, 
                                             false, 
                                             compositoresAtivos, 
                                             quantidadeDeArquivosPorCompositor);
            resultadoMelodia.setNomeCompositores(nomeCompositores);
            
            Resultado resultadoRitmo = new Resultado("Avaliacao_V3-Ritmo",
                                             Compositor.getContexto(),
                                             false, 
                                             true, 
                                             compositoresAtivos, 
                                             quantidadeDeArquivosPorCompositor);
            resultadoRitmo.setNomeCompositores(nomeCompositores);
            
            System.out.println("\nIniciando Avaliacao V3 com contexto "+c);
            
            //inicio do classificador
            for(int j = 0; j < quantidadeDeArquivosPorCompositor; j++){
                //preenche a lista com os arquivos que serao classificados
                for(int i = 0, k = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos()){
                        obrasAClassificar[k] = listaCompositores.get(i).getListaDeArquivos().removeFirst();
                        k++;
                    }
                }
                //classifica cada arquivo em cada compositor
                // a = index arquivo atual
                for(int a = 0; a < obrasAClassificar.length; a++){
                    System.out.println("\nClassificando: "+obrasAClassificar[a].getArquivoMidi().getName());

                    LinkedList<ItemResultado> resultadoObraAtualMelodia =  new LinkedList<ItemResultado>();
                    resultadoObraAtualMelodia.add(new ItemResultado(a));
                    LinkedList<ItemResultado> resultadoObraAtualRitmo =  new LinkedList<ItemResultado>();
                    resultadoObraAtualRitmo.add(new ItemResultado(a));

                    // m = index do compositor atual
                    for(int m = 0; m < listaCompositores.size(); m++){
                        if(listaCompositores.get(m).isAtivo() && listaCompositores.get(m).isComArquivos()){
                            listaCompositores.get(m).gerarModeloPPM(1);
                            double taxaMelodia = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), 1);
                            resultadoObraAtualMelodia.addLast(new ItemResultado(m));
                            resultadoObraAtualMelodia.getLast().setValor(taxaMelodia);

                            listaCompositores.get(m).gerarModeloPPM(2);
                            double taxaRitmo = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), 2);
                            resultadoObraAtualRitmo.addLast(new ItemResultado(m));
                            resultadoObraAtualRitmo.getLast().setValor(taxaRitmo);
                        }
                    }
                    resultadoMelodia.inserirResultado(resultadoObraAtualMelodia);
                    resultadoRitmo.inserirResultado(resultadoObraAtualRitmo);
                }
                for(int i = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos())
                        listaCompositores.get(i).getListaDeArquivos().addLast(obrasAClassificar[i]);
                }
            }

            // implementacao da classificacao do tipo 4

            Resultado resultadoFinal = new Resultado("Avaliacao_V3",
                                                 Compositor.getContexto(),
                                                 true, 
                                                 true, 
                                                 compositoresAtivos, 
                                                 quantidadeDeArquivosPorCompositor);
            resultadoFinal.setNomeCompositores(nomeCompositores);

            resultadoMelodia.ordenarResultados();
            resultadoMelodia.trocarTaxaPorPosicaoNoRanking();
            resultadoMelodia.ordenarCompositoresResultados();
            
            resultadoRitmo.ordenarResultados();
            resultadoRitmo.trocarTaxaPorPosicaoNoRanking();
            resultadoRitmo.ordenarCompositoresResultados();

            // faz a ordenacao dos resultados apos o a media do ranking obtido
            for(int j = 0; j < resultadoMelodia.getResultados().length; j++){
                LinkedList<ItemResultado> resultadoObraAtualTemp = new LinkedList<ItemResultado>();
                ItemResultado itemTemp = resultadoMelodia.getResultados()[j][0];
                resultadoObraAtualTemp.add(itemTemp);
                
                for(int r = 1; r <= resultadoMelodia.getQuantidadeDeCompositores(); r++){
                    itemTemp = new ItemResultado(resultadoMelodia.getResultados()[j][r].getCompositorItem());
                    itemTemp.setValor( ( resultadoMelodia.getResultados()[j][r].getValor() + 
                                         resultadoRitmo.getResultados()[j][r].getValor()  ) / 2);
                    resultadoObraAtualTemp.add(itemTemp);
                }

                resultadoFinal.inserirResultado(resultadoObraAtualTemp);
            }
            resultadoFinal.ordenarResultados();
            resultadoFinal.calcularAcertos();
            listaResultadosAV4.add(resultadoFinal);

        }//fim do for para gerar resultados
        
        return true;
    }
    
    /**
     * AV4 - Realiza uma comparacao das diferencas entre o primeiro e o segundo lugar
     * na classificacao em ritmo e também em melodia de maneira que a diferenca 
     * maior representa o resultado com maior certeza.
     * 
     * Realiza a avaliacao do arquivo utilizando o PPM e considerando cada arquivo
     * da base de dados como um modelo a fim de encontrar os mais semelhantes.
     * O contextoInicial e contextoFinal serve para programar as avaliacoes 
     * que serao realizadas. Para que seja realizada apenas uma avaliacao em um 
     * contexto especifico, o contexto inicial deve ser igual ao final.
     * 
     * @param contextoInicial
     * @param contextoFinal
     * @return
     * @throws IOException
     */
    public boolean avaliacaoV4(int contextoInicial, int contextoFinal) throws IOException{
        if(listaCompositores.size() == 0){
            System.out.println("Nao existem Compositores para realizar a classificacao");
            return false;
        }
        
        boolean compositoresComModelosEqualizados = true;
        int quantidadeDeArquivosPorCompositor = listaCompositores.getFirst().getListaDeArquivos().size();
        int compositoresAtivos = 0;
        Obra[] obrasAClassificar;
        String[] nomeCompositores;
        
        // verificacao especial para a classificacao v4
        if(quantidadeDeArquivosPorCompositor < 2){
            System.out.println("Nao existem arquivos suficientes para realizar a classificacao");
            return false;
                 
        }

        for(int i = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                compositoresAtivos++;
                if(listaCompositores.get(i).getListaDeArquivos().size() != quantidadeDeArquivosPorCompositor){
                    compositoresComModelosEqualizados = false;
                    System.out.println("Compositor "+listaCompositores.get(i).getNome()+" apresenta-se com quantidade diferente de arquivos!");
                }//fim if listaCompositores completos
            }//fim if listaCompositores ativos
        }//fim for lista de arquivos

        if(compositoresAtivos == 0 || compositoresComModelosEqualizados == false){
            System.out.println("Nao foi possivel classificar: Nenhum compositor ativo ou algum modelo incompativel!");
            return false;
        }
        
        obrasAClassificar = new Obra[compositoresAtivos];
        nomeCompositores = new String[compositoresAtivos];
        
        for(int i = 0, j = 0; i < listaCompositores.size(); i++){
            if(listaCompositores.get(i).isAtivo()){
                nomeCompositores[j] = Integer.toString(j).concat(":").concat(listaCompositores.get(i).getNome());
                j++;
            }
        }
        
        if(contextoInicial < 0){
            contextoInicial = 0;
        }
        if(contextoFinal < contextoInicial){
            contextoFinal = contextoInicial;
        }

        listaResultadosAV3 = new LinkedList<Resultado>();
        
        for(int c = contextoInicial; c <= contextoFinal; c++){ //realiza validacao cruzada com contextos determinados
            Compositor.setContexto(c);
            
            Resultado resultadoMelodia = new Resultado("Avaliacao_V4-Melodia",
                                             Compositor.getContexto(),
                                             true, 
                                             false, 
                                             compositoresAtivos, 
                                             quantidadeDeArquivosPorCompositor);
            resultadoMelodia.setNomeCompositores(nomeCompositores);
            
            Resultado resultadoRitmo = new Resultado("Avaliacao_V4-Ritmo",
                                             Compositor.getContexto(),
                                             false, 
                                             true, 
                                             compositoresAtivos, 
                                             quantidadeDeArquivosPorCompositor);
            resultadoRitmo.setNomeCompositores(nomeCompositores);
            
            System.out.println("\nIniciando Avaliacao V4 com contexto "+c);
            
            //inicio do classificador
            for(int j = 0; j < quantidadeDeArquivosPorCompositor; j++){
                //preenche a lista com os arquivos que serao classificados
                for(int i = 0, k = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos()){
                        obrasAClassificar[k] = listaCompositores.get(i).getListaDeArquivos().removeFirst();
                        k++;
                    }
                }
                //classifica cada arquivo em cada compositor
                // a = index arquivo atual
                for(int a = 0; a < obrasAClassificar.length; a++){
                    System.out.println("\nClassificando: "+obrasAClassificar[a].getArquivoMidi().getName());

                    LinkedList<ItemResultado> resultadoObraAtualMelodia =  new LinkedList<ItemResultado>();
                    resultadoObraAtualMelodia.add(new ItemResultado(a));
                    LinkedList<ItemResultado> resultadoObraAtualRitmo =  new LinkedList<ItemResultado>();
                    resultadoObraAtualRitmo.add(new ItemResultado(a));

                    // m = index do compositor atual
                    for(int m = 0; m < listaCompositores.size(); m++){
                        if(listaCompositores.get(m).isAtivo() && listaCompositores.get(m).isComArquivos()){
                            listaCompositores.get(m).gerarModeloPPM(1);
                            double taxaMelodia = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), 1);
                            resultadoObraAtualMelodia.addLast(new ItemResultado(m));
                            resultadoObraAtualMelodia.getLast().setValor(taxaMelodia);

                            listaCompositores.get(m).gerarModeloPPM(2);
                            double taxaRitmo = comprimeObraComModeloDoCompositor(obrasAClassificar[a], listaCompositores.get(m), 2);
                            resultadoObraAtualRitmo.addLast(new ItemResultado(m));
                            resultadoObraAtualRitmo.getLast().setValor(taxaRitmo);
                        }
                    }
                    resultadoMelodia.inserirResultado(resultadoObraAtualMelodia);
                    resultadoRitmo.inserirResultado(resultadoObraAtualRitmo);
                }
                for(int i = 0; i < listaCompositores.size(); i++){
                    if(listaCompositores.get(i).isAtivo() && listaCompositores.get(i).isComArquivos())
                        listaCompositores.get(i).getListaDeArquivos().addLast(obrasAClassificar[i]);
                }
            }

            resultadoMelodia.ordenarResultados();
            resultadoRitmo.ordenarResultados();

            Resultado resultadoFinal = new Resultado("Avaliacao_V4",
                                                 Compositor.getContexto(),
                                                 true, 
                                                 true, 
                                                 compositoresAtivos, 
                                                 quantidadeDeArquivosPorCompositor,
                                                 2);
            resultadoFinal.setNomeCompositores(nomeCompositores);

            // implementacao da avaliacao v4:
            // realiza uma comparacao das diferencas entre o primeiro e o segundo lugar
            // de maneira que a relacao que for de maior valor, indica maior certeza!
            for(int j = 0; j < resultadoMelodia.getResultados().length; j++){
                LinkedList<ItemResultado> resultadoObraAtualTemp = new LinkedList<ItemResultado>();

                double taxaMelodia, taxaRitmo;

                taxaMelodia = (resultadoMelodia.getResultados()[j][1].getValor() - 
                        resultadoMelodia.getResultados()[j][2].getValor()) / 
                        resultadoMelodia.getResultados()[j][2].getValor();

                taxaRitmo = (resultadoRitmo.getResultados()[j][1].getValor() - 
                        resultadoRitmo.getResultados()[j][2].getValor()) / 
                        resultadoRitmo.getResultados()[j][2].getValor();

                if(taxaMelodia < taxaRitmo){
                    resultadoObraAtualTemp.add(resultadoMelodia.getResultados()[j][0]);
                    resultadoObraAtualTemp.add(resultadoMelodia.getResultados()[j][1]);
                }else{
                    resultadoObraAtualTemp.add(resultadoRitmo.getResultados()[j][0]);
                    resultadoObraAtualTemp.add(resultadoRitmo.getResultados()[j][1]);
                }

                resultadoFinal.inserirResultado(resultadoObraAtualTemp);
            }
            resultadoFinal.calcularAcertos();
            listaResultadosAV3.add(resultadoFinal);
            
        }//fim do for para gerar resultados
        
        return true;
    }
    
    /**
     * Retorna os listaCompositores do classificador
     * @return the listaCompositores
     */
    public LinkedList<Compositor> getListaCompositores() {
        return listaCompositores;
    }

    /**
     * @return the listaResultadosAV1M
     */
    public LinkedList<Resultado> getListaResultadosAV1M() {
        return listaResultadosAV1M;
    }
    
    /**
     * @return the listaResultadosAV1R
     */
    public LinkedList<Resultado> getListaResultadosAV1R() {
        return listaResultadosAV1R;
    }

    /**
     * @return the listaResultadosAV2
     */
    public LinkedList<Resultado> getListaResultadosAV2() {
        return listaResultadosAV2;
    }
    
    /**
     * @return the listaResultadosAV2
     */
    public LinkedList<Resultado> getListaResultadosAV3() {
        return listaResultadosAV3;
    }

    /**
     * @return the listaResultadosAV4
     */
    public LinkedList<Resultado> getListaResultadosAV4() {
        return listaResultadosAV4;
    }
}
