/*
 * Esta classe representa um Compositor a ser utilizado com o PPM pelo classificador
 * e define todas as caracteristicas dos Compositores e suas obras. Cada Compositor
 * deve apresentar um valor para o contexto, um nome e uma lista de arquivos midi.
 * O classificador de Compositores ira realizar comparacoes entre os Compositores se
 * baseando nos arquivos de todos os Compositores.
 */

package classificador;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import ppm.PPMModel;

/**
 *
 * @author dj
 */
public class Compositor {

    /**
     * Nome do Compositor
     */
    private String nome;

    /**
     * Define se o Compositor podera ser utilizado em uma classificacao
     */
    private boolean ativo;

    /**
     * Define se a lista de obras do Compositor esta com ao menos um arquivo
     */
    private boolean comArquivos;

    /**
     * Lista de arquivos do Compositor
     */
    private LinkedList<Obra> listaDeArquivos = new LinkedList<Obra>();

    /**
     * Modelo PPM utilizado pelo Compositor para avaliar melodias
     */
    private PPMModel modeloPPMMelodia;
    
    /**
     * Modelo PPM utilizado pelo Compositor para avaliar ritmos
     */
    private PPMModel modeloPPMRitmo;
    
    /**
     * Modelo PPM utilizado pelo Compositor para avaliar melodias e ritmos
     */
    private PPMModel modeloPPMMelodiaRitmo;

    /**
     * Contexto a ser utilizado pelo PPM
     */
    private static int contexto = 0;
    
    /**
     * Contador de Compositores existentes para classificacao
     */
    private static int numeroDeCompositores = 0;

    /**
     * Cria o Compositor com informacoes padrao 
     */
    Compositor(String nome){
        numeroDeCompositores++;
        this.ativo = true;
        this.nome = nome;
    }

    /**
     * Cria o Compositor considerando um nome padrao
     */
    Compositor() {
        numeroDeCompositores++;
        this.ativo = true;
        this.nome = "Compositor "+numeroDeCompositores;
    }

    /**
     * Adiciona um arquivo na lista de arquivos do Compositor. Apos adicionar o
     * arquivo e' preciso atualizar o Modelo PPM para incluir as caracteristicas
     * do novo arquivo.
     * @param midi
     */
    public void adicionarArquivo(File midi){
        System.out.println("\nAdicionando arquivo "+midi.getName()+" na lista de obras do Compositor "+this.nome);
        getListaDeArquivos().add(new Obra(midi));
        getListaDeArquivos().getLast().preencherListaDeNotas();
        comArquivos = true;
    }
    
    /**
     * Remove um arquivo da lista de arquivos do Compositor. Apos a remocao do
     * arquivo ha a necessidade de atualizar o Modelo PPM para atualizar o
     * modelo com os arquivos existentes.
     * @param idArquivo
     */
    public Obra removeArquivo(int idArquivo){
        String nomeArquivoRemovido = getListaDeArquivos().get(idArquivo).getArquivoMidi().getName();
        Obra removido = this.listaDeArquivos.remove(idArquivo);
        System.out.println("O arquivo "+nomeArquivoRemovido+" foi removido das obras de "+this.nome);
        if(getListaDeArquivos().size() == 0) {
            comArquivos = false;
        }
        return removido;
    }

    
    /**
     * Gera o Modelo PPM baseado nos intervalos
     * 
     * @throws IOException
     */
    public void gerarModeloPPM(int tipoDeModelo) throws IOException{
        int quantidadeDeArquivos = this.listaDeArquivos.size();
        
        if(tipoDeModelo == 0){
            System.out.println("  Gerando Modelo PPM Melodico e Ritmico para o Compositor "+this.nome);
            this.modeloPPMMelodiaRitmo = new PPMModel(Compositor.getContexto());
        }else
        if(tipoDeModelo == 1){
            System.out.println("  Gerando Modelo PPM Melodico para o Compositor "+this.nome);
            this.modeloPPMMelodia = new PPMModel(Compositor.getContexto());
        }else{
            System.out.println("  Gerando Modelo PPM Ritmico para o Compositor "+this.nome);
            this.modeloPPMRitmo = new PPMModel(Compositor.getContexto());
        }
        
       
        for(int i = 0; i < quantidadeDeArquivos; i++){
            this.listaDeArquivos.get(i).gerarListaDeArticulacoes(tipoDeModelo);
            if(getListaDeArquivos().get(i).getListaDeArticulacoes(tipoDeModelo).size() > 0){
                LinkedList<Integer> intervalos = getListaDeArquivos().get(i).getListaDeArticulacoes(tipoDeModelo);
                for(Integer intervalo : intervalos){
                    if(tipoDeModelo == 0){
                        this.modeloPPMMelodiaRitmo.encode(intervalo, true);
                    }else
                    if(tipoDeModelo == 1){
                        this.modeloPPMMelodia.encode(intervalo, true);                        
                    }else{
                        this.modeloPPMRitmo.encode(intervalo, true);
                    }
                    
                }
            }
        }  
    }
    

    /**
     * Checa a menor quantidade de notas dos arquivos do Compositor comparando
     * todos os arquivos existentes na sua lista de arquivos.
     * @return a menor quantidade de notas dos arquivos do Compositor
     */
    public int checarMenorQuantidadeDeIntervalos(){
        int menorQuantidadeDeIntervalos = Integer.MAX_VALUE;
        int tempLimite = Obra.getLimiteDeNotas();
        for (Obra arquivoMidi : this.listaDeArquivos) {
            // utiliza-se o zero para considerar o arquivo por completo na hora da checagem
            Obra.setLimiteDeNotas(0);
            // gera a lista de intervalos melodicos por ser mais veloz
            arquivoMidi.gerarListaDeArticulacoes(1);

            if(menorQuantidadeDeIntervalos > arquivoMidi.getQuatidadeDeArticulacoes()){
                menorQuantidadeDeIntervalos = arquivoMidi.getQuatidadeDeArticulacoes();
            }
        }
        Obra.setLimiteDeNotas(tempLimite);
        System.out.println("Menor quantidade de intervalos dentre as obras do "
                + "Compositor "+this.nome+" : "+menorQuantidadeDeIntervalos);

        return menorQuantidadeDeIntervalos;
    }

    /**
     * Retorna o nome do Compositor
     * @return the nome
     */
    public String getNome() {
        return nome;
    }

    /**
     * Define um novo nome para o Compositor
     * @param nome the nome to set
     */
    public void setNome(String nome) {
        this.nome = nome;
    }

    /**
     * Retorna o contexto utilizado pelo modelo PPM do Compositor
     * @return the contexto
     */
    public static int getContexto() {
        return contexto;
    }

    /**
     * Define o novo contexto a ser utilizado para tratamento das obras que
     * fazem parte do modelo PPM do Compositor
     * @param contexto the contexto to set
     */
    public static void setContexto(int contextoGeral) {
        contexto = contextoGeral;
    }

    /**
     * Retorna o modelo PPM do Compositor
     * O tipo do modelo e' definido a partir de:
     * 0 = Melodia + Ritmo
     * 1 = Melodia
     * 2 = Ritmo
     * @return the modelo
     */
    public PPMModel getModelo(int tipoDoModelo) {
        if(tipoDoModelo == 0){
            return modeloPPMMelodiaRitmo;
        }else
        if(tipoDoModelo == 1){
            return modeloPPMMelodia;
        }else{
            return modeloPPMRitmo;
        }
    }

    /**
     * Verifica se o Compositor esta ativo para classificacao de arquivos
     * @return the ativo
     */
    public boolean isAtivo() {
        return ativo;
    }

    /**
     * Define o Compositor como ativo para a classificacao de arquivos
     * @param ativo the ativo to set
     */
    public void setAtivo(boolean ativo) {
        this.ativo = ativo;
    }

    /**
     * Verifica se o Compositor contem arquivos para serem classificados
     * @return the comArquivos
     */
    public boolean isComArquivos() {
        return comArquivos;
    }

    /**
     * Retorna a lista de arquivos do Compositor
     * @return the listaDeArquivos
     */
    public LinkedList<Obra> getListaDeArquivos() {
        return listaDeArquivos;
    }

}
