/*
 * Classe representante das obras do Compositor no fomato midi
 *
 * Esta classe contem os metodos basicos para utilizacao dos arquivos.
 * Apresenta os metodos de leitura e decodificacao dos arquivos midi
 * para transforma-los em listas de notas.
 */

package classificador;

import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;

/**
 *
 * @author dj
 */
public class Obra {
    
    private File arquivoMidi;
    Sequence sequenciador;
    long quantidadeDeTicks;
    long duracao;
    int resolucao;
    float tipoDeDivisao;
    Track[] trilhas;

    /**
     * Lista das notas existentes no arquivo
     */
    private LinkedList<Nota> listaDeNotas = new LinkedList<Nota>();

    /*
     * lista de intervalos melodicos
     */
    private LinkedList<Integer> listaDeArticulacoesMelodicas = new LinkedList<Integer>();

    /*
     * lista de proporcoes ritmicas
     */
    private LinkedList<Integer> listaDeArticulacoesRitmicas = new LinkedList<Integer>();

    /*
     * lista de intervalos melodicos e proporcoes ritmicas (MelodicoRitmicas)
     */
    private LinkedList<Integer> listaDeArticulacoesMelodicoRitmicas = new LinkedList<Integer>();


    /**
     * Indica se ja foi gerada a lista de notas
     */
    private boolean comNotas = false;

    /**
     * Define se as harmonias existentes entre as notas serao removidas
     */
    private static boolean semHarmonias = true;

    /**
     * Considera os silencios existentes entre as notas
     */
    private static boolean comPausas = false;

    /**
     * Define a quantidade de notas a se considerar.
     * Caso o valor seja 0, todos os intervalos serao considerados
     */
    private static int limiteDeNotas = 0;

    /**
     * Construtor baseado no endereco fisico do arquivo
     * @param enderecoDoArquivo endereco fisico do arquivo midi
     */
    public Obra(File enderecoDoArquivo) {
        try {
            arquivoMidi = enderecoDoArquivo;
            sequenciador = MidiSystem.getSequence(arquivoMidi);
            quantidadeDeTicks = sequenciador.getTickLength();
            duracao = sequenciador.getMicrosecondLength();
            tipoDeDivisao = sequenciador.getDivisionType();
            resolucao = sequenciador.getResolution();
            trilhas = sequenciador.getTracks();
            System.out.println("Descricao do arquivo: " + arquivoMidi.getName());
            System.out.print("* ticks: " + quantidadeDeTicks);
            System.out.print(", duracao: " + duracao);
            System.out.print(", divisao: ");
            if (tipoDeDivisao == Sequence.PPQ) {
                System.out.print(", PPQ");
            } else {
                System.out.print(", SMPTE");
            }
            System.out.print(", resolucao: " + resolucao);
            System.out.println(", trilhas: " + trilhas.length);
        } catch (InvalidMidiDataException ex) {
            System.out.println("Execao: "+ex.getMessage().toString());
        } catch (IOException ex) {
            System.out.println("Execao: "+ex.getMessage().toString());
        }
    }

    /*
     * Preenche a lista de notas com todas as notas da trilha
     * No momento adiciona as notas em sequencia apos o fim de cada trilha
     */
    public void preencherListaDeNotas(){

        if(trilhas != null){

            listaDeNotas.clear();
            LinkedList<Nota> notasOff = new LinkedList<Nota>();

//            for(int i = 0; i < trilhas.length; i++){
            for(int i = 0; i < 1; i++){
                Track trilha = trilhas[ i ];
                for( int j = 0; j < trilha.size(); j++){
                    //Solicita o proximo evento da trilha
                    MidiEvent evento = trilha.get(j);
                    //Transforma o evento em mensagem
                    MidiMessage mensagem = evento.getMessage();
                    //Checa se a mensagem tem informacao de notas
                    if(mensagem instanceof ShortMessage){
                        ShortMessage mens = (ShortMessage) mensagem;
                        switch( mens.getCommand() ){
                            case 0x90: //Nota On
                                this.listaDeNotas.addLast(new Nota(mens.getData1(), evento.getTick()));
                                break;
                            case 0x80: //Nota Off
                                notasOff.addLast(new Nota(mens.getData1(), evento.getTick()));
                                break;
                        }//fim switch ShortMessage
                    }//fim if da chegagem por ShortMessage
                }//fim for da decodificacao de eventos do track
            }//fim for leitura de tracks

            completarListaDeNotas(notasOff);
        }//fim if checagem se o arquivo apresenta tracks
    }

    /**
     * Completa a lista de notas realizando comparacoes entre as listas de notas
     * @param notasOff
     */
    private void completarListaDeNotas(LinkedList<Nota> notasOff){
        boolean completouInfoNota = false;

        System.out.print("* notas on: "+listaDeNotas.size());
        System.out.println(", notas off: "+notasOff.size());
        for( int k = 0; k < listaDeNotas.size(); k++){
            for(int m = 0; m < notasOff.size(); m++){
                if(listaDeNotas.get(k).isEquals(notasOff.get(m))){
                    long tickFinal = notasOff.get(m).getTickInicial();
                    listaDeNotas.get(k).setTickFinal(tickFinal);
                    notasOff.remove(m);
                    completouInfoNota = true;
                    m = notasOff.size();
                }//fim if compara notas
            }//fim for m: lista de notasOff
            if(!completouInfoNota){
                listaDeNotas.get(k).setTickFinal(
                                listaDeNotas.get(k).getTickInicial());
                listaDeNotas.get(k).imprimirNotaInfo();
            }
        }//fim for k: listaDeNotas

        if(notasOff.size() > 0){
            System.out.println("* A lista de notas off ainda contem"
                    + "elementos!");
        }
        else{
            System.out.println("* Preenchimento completo da lista de "
                    + "notas realizado com sucesso!");
        }//fim if checa lista de notasOff

        this.comNotas = true;
    }

    /**
     * Imprime as informacoes das notas que fazem parte da lista de notas
     * @param infoCompleta define se se deseja as informacoes completas
     */
    public void imprimirListaDeNotas(boolean infoCompleta){
        if(!isComNotas()){
            this.preencherListaDeNotas();
        }
        if(infoCompleta){
            for(int i = 0; i < listaDeNotas.size(); i++){
                listaDeNotas.get(i).imprimirNotaInfo();
            }
        }
        else{
            for(int i = 0; i < listaDeNotas.size(); i++){
                listaDeNotas.get(i).imprimirNota();
            }
        }
    }
    
    
    public void gerarListaDeArticulacoes(int tipoDeModelo){
        if(tipoDeModelo == 0){
            gerarListaMelodicoRitmica();
        }else
        if(tipoDeModelo == 1){
            gerarListaDeIntervalosMelodicos();
        }else{
            gerarListaDeProporcoesRitmicas();
        }
    }
    

    /**
     * Gera a lista de intervalos melodicos existentes entre as notas do arquivo.
     */
    private void gerarListaDeIntervalosMelodicos(){

        if(listaDeNotas.size() == 0){
            this.preencherListaDeNotas();
            if(listaDeNotas.size() == 0) return;
        }

        LinkedList<Nota> listaDeNotasTemp = (LinkedList<Nota>) listaDeNotas.clone();

        if(listaDeNotasTemp.size() != 0){

            if(semHarmonias){
                removerHarmonias(listaDeNotasTemp);
            }

            if(comPausas){
                considerarPausas(listaDeNotasTemp);
            }

            // Cria a lista de intervalos
            if(listaDeNotasTemp.size() != 0){

                listaDeArticulacoesMelodicas.clear();

                // Define a quantidade de notas a serem lidas
                int notasAAnalisar = listaDeNotasTemp.size();
                if(limiteDeNotas != 0){
                    if(limiteDeNotas < notasAAnalisar){
                        notasAAnalisar = limiteDeNotas;
                    }
                }

                int temp = listaDeNotasTemp.getFirst().getNumeroDaNota();
                int valorFinal = 0;
                for(int i = 1; i < notasAAnalisar; i++){
                    // Verifica se encontrou um silencio antes de gerar o intervalo
                    if(listaDeNotasTemp.get(i).getNumeroDaNota() != Integer.MAX_VALUE ){
                        if(temp != 254){
                            valorFinal = ( 127 + listaDeNotasTemp.get(i).getNumeroDaNota() - temp);
                            if(valorFinal < 0) valorFinal = 0;
                            else if(valorFinal > 255) valorFinal = 255;
                            listaDeArticulacoesMelodicas.add(valorFinal);
                        }else{
                            valorFinal = temp;
                        }
                        temp = listaDeNotasTemp.get(i).getNumeroDaNota();

                    }else{
                        // 254 = pausa
                        temp = 254;
                        listaDeArticulacoesMelodicas.add(temp);
                    }
                }
//                System.out.println("* "+this.arquivoMidi.getName()+" : "+listaDeArticulacoes.size()+" intervalos melodicos.");

            }
            else{
                System.out.println("* Lista de intervalos vazia!");
            }
        }
        else{
            System.out.println("* Lista de notas vazia!");
        }
    }

    /**
     * Gera a lista de proporcoes ritmicas existentes entre as notas do arquivo.
     */
    private void gerarListaDeProporcoesRitmicas(){

        if(listaDeNotas.size() == 0){
            this.preencherListaDeNotas();
            if(listaDeNotas.size() == 0) return;
        }

        LinkedList<Nota> listaDeNotasTemp = (LinkedList<Nota>) listaDeNotas.clone();

        if(listaDeNotasTemp.size() != 0){

            if(semHarmonias){
                removerHarmonias(listaDeNotasTemp);
            }

            if(comPausas){
                considerarPausas(listaDeNotasTemp);
            }

            // Cria a lista de intervalos
            if(listaDeNotasTemp.size() != 0){

                listaDeArticulacoesRitmicas.clear();

                // Define a quantidade notas a analisar
                int notasAAnalisar = listaDeNotasTemp.size();
                if(limiteDeNotas != 0){
                    if(limiteDeNotas < notasAAnalisar){
                        notasAAnalisar = limiteDeNotas;
                    }
                }

                double temp = listaDeNotasTemp.getFirst().getTicks();
                int valorFinal = 0;
                for(int i = 1; i < notasAAnalisar; i++){

                    if(listaDeNotasTemp.get(i).getTicks() != 0){
                        double valor = (double) ( (double)listaDeNotasTemp.get(i).getTicks()/temp );
                        if(valor >= 1.0){
                            valorFinal = ( (int) ((double) (127 + valor*10.0)));
                        }else{
                            valorFinal = ( (int) ((double) (127 - valor*100.0)));
                        }
                        if(valorFinal < 0) valorFinal = 0;
                        else if(valorFinal > 256) valorFinal = 256;
                        listaDeArticulacoesRitmicas.add(valorFinal);
                    }
                    temp = listaDeNotasTemp.get(i).getTicks();
                }
            }
            else{
                System.out.println("* Lista de proporcoes ritmicas vazia!");
            }
        }
        else{
            System.out.println("* Lista de notas vazia!");
        }
    }

    /**
     * Gera a lista de intervalos melodicos juntos com as proporcoes ritmicas
     * existentes entre as notas do arquivo.
     * 
     */
    private void gerarListaMelodicoRitmica(){

        if(listaDeNotas.size() == 0){
            this.preencherListaDeNotas();
            if(listaDeNotas.size() == 0) return;
        }

        LinkedList<Nota> listaDeNotasTemp = (LinkedList<Nota>) listaDeNotas.clone();

        if(listaDeNotasTemp.size() != 0){

            if(semHarmonias){
                removerHarmonias(listaDeNotasTemp);
            }

            if(comPausas){
                considerarPausas(listaDeNotasTemp);
            }

            // Cria a lista de intervalos
            if(listaDeNotasTemp.size() != 0){

                listaDeArticulacoesMelodicoRitmicas.clear();

                // Define a quantidade notas a analisar
                int notasAAnalisar = listaDeNotasTemp.size();
                if(limiteDeNotas != 0){
                    if(limiteDeNotas < notasAAnalisar){
                        notasAAnalisar = limiteDeNotas;
                    }
                }


                //TODO: metodo para armazenar informacoes de ritmo e melodia
                // usar listaDeArticulacoesMelodicoRitmicas;

            }
            else{
                System.out.println("* Lista de proporcoes ritmicas vazia!");
            }
        }
        else{
            System.out.println("* Lista de notas vazia!");
        }
    }

    /*
     * Se for solicitado remover harmonias (notas executadas no mesmo instante)
     * sera realizada uma varredura na lista de notas removendo os pares harmonicos
     */
    private void removerHarmonias(LinkedList<Nota> lista){
        boolean removerOPrimeiro = false;
        for(int i = 0; (i+1) < lista.size(); ){
            if(lista.get(i).getTickInicial() == lista.get(i+1).getTickInicial()){
                lista.remove(i+1);
                removerOPrimeiro = true;
            }else{
                i++;
                if(removerOPrimeiro){
                    lista.remove(i);
                    removerOPrimeiro = false;
                }
            }
        }
    }

    /*
     * Se existir silencio entre duas notas, sera adicionada uma nota
     * referente a este silencio
     */
    private void considerarPausas(LinkedList<Nota> lista){
        
        Nota notaTemp2 = lista.getFirst();
        long tickInicial = notaTemp2.getTickFinal();
        long tickFinal = 0;
        for(int i = 1; i < lista.size(); i++){
            tickFinal = lista.get(i).getTickInicial();
            if(tickInicial < tickFinal){
                // a nova nota sera criada com o valor de silencio igual a Integer.MAX_VALUE
                Nota novaNota = new Nota(Integer.MAX_VALUE, tickInicial);
                novaNota.setTickFinal(tickFinal);
                lista.add(i,novaNota);
                tickInicial = tickFinal;
            }
            else{
                tickInicial = lista.get(i).getTickFinal();
            }
        }
    }

    /*
     * Imprimir lista de articulacoes
     */
    public void imprimirListasDeArticulacoes(){
        System.out.println("Arquivo: "+this.arquivoMidi.getName());
        if(!isComNotas()){
            this.preencherListaDeNotas();
        }

        this.gerarListaDeIntervalosMelodicos();
        this.gerarListaDeProporcoesRitmicas();
        this.gerarListaMelodicoRitmica();

        System.out.println("Lista de articulacoes Melodicas:");
        for(Integer articulacao : listaDeArticulacoesMelodicas){
            System.out.print(articulacao+" ");
        }
        System.out.println("");

        System.out.println("Lista de articulacoes Ritmicas:");
        for(Integer articulacao : listaDeArticulacoesRitmicas){
            System.out.print(articulacao+" ");
        }
        System.out.println("");

        System.out.println("Lista de articulacoes MelodicoRitmicas:");
        for(Integer articulacao : listaDeArticulacoesMelodicoRitmicas){
            System.out.print(articulacao+" ");
        }
        System.out.println("");
    }

    /**
     * @return the listaDeArticulacoesMelodicas
     */
    public LinkedList<Integer> getListaDeArticulacoes(int tipoDeModelo) {
        if(tipoDeModelo == 0){
            return listaDeArticulacoesMelodicoRitmicas;
        }else
        if(tipoDeModelo == 1){
            return listaDeArticulacoesMelodicas;
        }else{
            return listaDeArticulacoesRitmicas;
        }
    }


    /**
     * Retorna a quantidade de articulacoes da obra. Utiliza da lista de
     * articulacoes melodicas, visto que todas as listas de articulacao terao o
     * mesmo tamanho.
     * @return the listaDeArticulacoesMelodicas.size();
     */
    public int getQuatidadeDeArticulacoes() {
        if(listaDeArticulacoesMelodicas.size() == 0){
            gerarListaDeArticulacoes(1);
        }
        return listaDeArticulacoesMelodicas.size();
    }

    /**
     * @return the arquivoMidi
     */
    public File getArquivoMidi() {
        return arquivoMidi;
    }

    /**
     * @return the semHarmonias
     */
    public static boolean isSemHarmonias() {
        return semHarmonias;
    }

    /**
     * @param aRemoverHarmonias the semHarmonias to set
     */
    public static void setSemHarmonias(boolean aRemoverHarmonias) {
        semHarmonias = aRemoverHarmonias;
    }

    /**
     * @return the considerarPausas
     */
    public static boolean isComPausas() {
        return comPausas;
    }

    /**
     * @param aConsiderarPausas the considerarPausas to set
     */
    public static void setComPausas(boolean aConsiderarPausas) {
        comPausas = aConsiderarPausas;
    }

    /**
     * @return the quantidadeDeIntervalos
     */
    public static int getLimiteDeNotas() {
        return limiteDeNotas;
    }

    /**
     * @param aQuantidadeDeIntervalos the quantidadeDeIntervalos to set
     */
    public static void setLimiteDeNotas(int aLimiteDeNotas) {
        limiteDeNotas = aLimiteDeNotas;
    }

    /**
     * Checa se a lista de notas da obra ja foi criada
     * @return the comNotas
     */
    public boolean isComNotas() {
        return comNotas;
    }
}